local errType = require 'errType'
local utils = require 'utils'

local FileSystem_SimpleList = require 'FileSystem_SimpleList'

local Network_GetState = {}

function Network_GetState.hwaddr2name(addr)
    local name = '!'
    if addr then
        for i = 1, #addr do
            local c = addr:sub(i, i)
            if c ~= ':' then
                name = name .. c
            end
        end
    end
    return name
end

function readMatchSection(config, line)
    local name = line:match('^%s*Name%s*=%s*(%w+)%s*$')
    if name then
        config.name = name
        return
    end
    local mac = line:match('^%s*MACAddress%s*=%s*([%w|:]+)%s*$')
    if mac then
        config.hwaddr = mac
        return
    end
end

function readNetworkSection(config, line)
    local dhcpClient = line:match('^%s*DHCP%s*=%s*(.*)%s*$')
    if dhcpClient then
        if dhcpClient == 'ipv4' or dhcpClient == 'ipv6' or dhcpClient == 'yes' then
            config.dhcpClient = true
        else
            config.dhcpClient = false
        end
        return
    end
    local dhcpServer = line:match('^%s*DHCPServer%s*=%s*(%w+)%s*$')
    if dhcpServer then
        if dhcpServer == 'yes' then
            config.dhcpServer = true
        else
            config.dhcpServer = false
        end
        return
    end
    local addr = line:match('^%s*Address%s*=%s*(.+)%s*$')
    if addr then
        if config.addresses == nil then
            config.addresses = {}
        end
        table.insert(config.addresses, addr)
        return
    end
    local gateway = line:match('^%s*Gateway%s*=%s*(.+)%s*$')
    if gateway then
        if config.gateways == nil then
            config.gateways = {}
        end
        table.insert(config.gateways, gateway)
        return
    end
    local dns = line:match('^%s*DNS%s*=%s*(.+)%s*$')
    if dns then
        if config.dns == nil then
            config.dns = {}
        end
        table.insert(config.dns, dns)
        return
    end
    local ipmasq = line:match('^%s*IPMasquerade%s*=%s*(%w+)%s*$')
    if ipmasq then
        if ipmasq == 'yes' then
            config.ipMasquerade = true
        else
            config.ipMasquerade = false
        end
        return
    end
    local ip6ra = line:match('^%s*IPv6AcceptRA%s*=%s*(%w+)%s*$')
    if ip6ra then
        if ip6ra == 'yes' then
            config.ipv6AcceptRA = true
        else
            config.ipv6AcceptRA = false
        end
        return
    end
end

function parseConfigFile(path)
    local config = {}
    local f = io.open(path, "r")
    if f ~= nil then
        local content = f:read("*all")
        f:close()
        
        local insideMatchSection = false
        local insideNetworkSection = false
        local lines = utils.split(content, '\n')
        for i, line in ipairs(lines) do
            if line:match('%s*%[%s*Match%s*%]%s*') then
                insideMatchSection = true
                insideNetworkSection = false
            elseif line:match('%s*%[%s*Network%s*%]%s*') then
                insideNetworkSection = true
                insideMatchSection = false
            end
            if insideMatchSection then
                readMatchSection(config, line)
            elseif insideNetworkSection then
                readNetworkSection(config, line)
            end
        
        end
    end
    return config
end

function readNetDevSection(config, line)
    local kind = line:match('^%s*Kind%s*=%s*(%w+)%s*$')
    if kind then
        config.kind = kind
        return
    end
end

function parseDeviceFile(path)
    local config = {}
    local f = io.open(path, "r")
    if f ~= nil then
        local content = f:read("*all")
        f:close()
        
        local insideDevSection = false
        local lines = utils.split(content, '\n')
        for i, line in ipairs(lines) do
            if line:match('%s*%[%s*NetDev%s*%]%s*') then
                insideDevSection = true
            end
            if insideDevSection then
                readNetDevSection(config, line)
                
                -- read name or mac
                readMatchSection(config, line)
            end
        
        end
    end
    return config
end

function Network_GetState.invoke(...)
    
    local stdout, err, errMsg = utils.shell({'ip', '-o', '-s', 'link'})
    if err ~= nil then
        return nil, err, errMsg
    end
    
    local ifaces = {}
    
    local linkLines = utils.split(stdout, '\n')
    for key, line in pairs(linkLines) do
        local id, name, flags = string.match(line,
            "(%d+):%s+(%w+):%s*%<(.+)%>")
        
        local mac = string.match(line,
            "link/%w+%s+([%w|:]+)")
        
        local rxBytes, rxPackets, rxErrors, rxDropped, rxOverrun, rxMcast = string.match(line,
            "RX:%s+bytes%s+packets%s+errors%s+dropped%s+overrun%s+mcast%s*\\" ..
            "%s*(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+).*")
        
        local txBytes, txPackets, txErrors, txDropped, txCarrier, txColls = string.match(line,
            "TX:%s+bytes%s+packets%s+errors%s+dropped%s+carrier%s+collsns%s*\\" ..
            "%s*(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+).*")
        
        if id then
            id = tonumber(id)
            local iface
            for i, ifc in ipairs(ifaces) do
                if ifc.id == id then
                    iface = ifc
                end
            end
            
            if iface == nil then
                iface = {}
                iface.name = name
                iface.id = id
                table.insert(ifaces, iface)
            end
            
            iface.hwaddr = mac
            iface.rx = {}
            iface.rx.bytes = rxBytes
            iface.rx.packets = rxPackets
            iface.rx.errors = rxErrors
            iface.rx.dropped = rxDropped
            iface.tx = {}
            iface.tx.bytes = txBytes
            iface.tx.packets = txPackets
            iface.tx.errors = txErrors
            iface.tx.dropped = txDropped
            
            flags = utils.split(flags, ',')
            iface.up = false
            iface.plugged = false
            for i, v in ipairs(flags) do
                if v == 'UP' then
                    iface.up = true
                end
                if v == 'LOWER_UP' then
                    iface.plugged = true
                end
            end
        end
    end
    
    local stdout, err, errMsg = utils.shell({'ip', '-o', 'addr'})
    if err ~= nil then
        return nil, err, errMsg
    end
    
    local lines = utils.split(stdout, '\n')
    
    for key, line in pairs(lines) do
        local id, name, family, ip, scope, flags, validLft, prefferedLft = string.match(line,
            "(%d+):%s+(%w+)%s+(%w+)%s+([%w|.|:|%/]+)%s+scope%s*(%w+)(.*)\\%s+valid%_lft%s+(%w+)%s+preferred%_lft%s+(%w+)")
        
        if id == nil then
            id, name, family, ip, brd, scope, flags, validLft, prefferedLft = string.match(line,
                "(%d+):%s+(%w+)%s+(%w+)%s+([%w|.|:|%/]+)%s+brd%s+([%w|.|:|%/]+)%s+scope%s+(%w+)(.*)\\%s+valid%_lft%s+(%w+)%s+preferred%_lft%s+(%w+)")
        end
        
        if id ~= nil then
            local addr = {}
            addr.family = utils.trim(family)
            addr.ip = utils.trim(ip)
            addr.broadcast = utils.trim(brd)
            addr.scope = utils.trim(scope)
            addr.lifetime = {}
            addr.lifetime.validLft = utils.trim(validLft)
            addr.lifetime.prefferedLft = utils.trim(prefferedLft)
            
            if flags:match('dynamic') then
                addr.dynamic = true
            else
                addr.dynamic = false
            end
            
            id = tonumber(id)
            local iface
            for i, ifc in ipairs(ifaces) do
                if ifc.id == id then
                    iface = ifc
                end
            end
            
            if iface == nil then
                iface = {}
                iface.name = name
                iface.id = id
                table.insert(ifaces, iface)
            end
            
            if iface.addresses == nil then
                iface.addresses = {}
            end
            table.insert(iface.addresses, addr)
        end
    end
    
    -- process *linkLines* again to match masters and slaves together
    for key, line in pairs(linkLines) do
        local id, master = string.match(line, "(%d+):.*master%s+(%w+)")
        if id then
            id = tonumber(id)
            local iface
            for i, ifc in ipairs(ifaces) do
                if ifc.id == id then
                    iface = ifc
                end
            end
            
            if iface == nil then
                iface = {}
                iface.name = name
                iface.id = id
                table.insert(ifaces, iface)
            end
            
            iface.master = master
            for k, ifc in pairs(ifaces) do
                if ifc.name == master then
                    if ifc.slaves == nil then
                        ifc.slaves = {}
                    end
                    table.insert(ifc.slaves, iface.name)
                end
            end
        end
    end
    
    local fsItems, err, errMsg = FileSystem_SimpleList.invoke({"/etc/systemd/network"})
    if fsItems == null then
        return nil, errType.codes.InternalError, 'Failed to query network configuration files. ' .. errMsg
    end
    
    -- parse *.netdev files
    for i, fsItem in ipairs(fsItems) do
        if utils.endsWith(fsItem, '.netdev') then
            local config = parseDeviceFile('/etc/systemd/network/' .. fsItem)
            local ifaceFound = false
            for i, iface in ipairs(ifaces) do
                if (config.name and iface.name == config.name) or (config.hwaddr and iface.hwaddr == config.hwaddr) then
                    iface.device = config
                    ifaceFound = true
                    break
                end
            end
            
            if ifaceFound == false then
                local iface = {}
                iface.name = config.name
                iface.hwaddr = config.hwaddr
                iface.device = config
                iface.inactive = true
                table.insert(ifaces, iface)
            end
        end
    end
    
    -- parse *.network files
    for i, fsItem in ipairs(fsItems) do
        if utils.endsWith(fsItem, '.network') then
            local config = parseConfigFile('/etc/systemd/network/' .. fsItem)
            local ifaceFound = false
            for i, iface in ipairs(ifaces) do
                if (config.name and iface.name == config.name) or (config.hwaddr and iface.hwaddr == config.hwaddr) then
                    iface.config = config
                    iface.managed = true
                    if config.hwaddr and iface.hwaddr == config.hwaddr then
                        iface.managedByHwAddr = true
                    end
                    ifaceFound = true
                    break
                end
            end
            
            if ifaceFound == false then
                local iface = {}
                iface.name = config.name
                iface.hwaddr = config.hwaddr
                iface.config = config
                iface.inactive = true
                iface.managed = true
                if config.hwaddr then
                    iface.managedByHwAddr = true
                end
                table.insert(ifaces, iface)
            end
        end
    end
    
    return ifaces
end

return Network_GetState
