local errType = require 'errType'
local utils = require 'utils'
local ipValidator = require 'ipAddrValidator';
local Network_GetState = require 'Network_GetState'
local removeMethod = require 'FileSystem_Remove'

local Network_SetConfig = {}

function checkIPsWithMask(addrs)
    for i, addr in ipairs(addrs) do
        local res = ipValidator.getIPType(addr)
        if res == ipValidator.Type.ERROR or res == ipValidator.Type.STRING then
            return 'Argument \'' .. tostring(addr) .. '\' is not valid IPv4 or IPv6 address.'
        end
        if res == ipValidator.Type.IPv4 or res == ipValidator.Type.IPv6 then
            return 'Argument \'' .. tostring(addr) .. '\' has no mask specified.'
        end
    end
end

function checkIPsWithoutMask(addrs)
    for i, addr in ipairs(addrs) do
        local res = ipValidator.getIPType(addr)
        if res == ipValidator.Type.ERROR or res == ipValidator.Type.STRING then
            return nil, errType.codes.InvalidParams, 'Argument \'' .. tostring(addr) .. '\' is not valid IPv4 or IPv6 address.'
        end
        if res == ipValidator.Type.IPv4_withMask or res == ipValidator.Type.IPv6_withMask then
            return nil, errType.codes.InvalidParams, 'Argument \'' .. tostring(addr) .. '\' must not have mask specified.'
        end
    end
end

function checkBooleanArgument(arg)
    return arg == nil or arg == true or arg == false
end

function checkInt(arg)
    if arg == nil or type(arg) ~= 'number' then
        return false
    end
    if math.floor(arg) == arg then
        return true
    else
        return false
    end
end

function bool2ConfigValue(val)
    if val then
        return 'yes'
    else
        return 'no'
    end
end

function restartNetworkService()
    local stdout, err, errMsg = utils.shell({'systemctl', 'restart', 'systemd-networkd'})
    if err ~= nil then
        return false
    end
    return true
end

function flushInterface(iface)
    local stdout, err, errMsg = utils.shell({'ip', 'addr', 'flush', 'dev', iface})
    if err ~= nil then
        return false
    end
    return true
end

function Network_SetConfig.invoke(args)
    local interfaceName, hwaddr, staticAddresses, dnsServers, gateways,
        ipv6AcceptRA, ipMasquerade, dhcpClient,
        dhcpServer, dhcpServerPoolOffset, dhcpServerPoolSize
    
    if args ~= nil then
        if utils.isArray(args) then
            interfaceName = args[1]
            hwaddr = args[2]
            staticAddresses = args[3]
            dnsServers = args[4]
            gateways = args[5]
            ipv6AcceptRA = args[6]
            ipMasquerade = args[7]
            dhcpClient = args[8]
            dhcpServer = args[9]
            dhcpServerPoolOffset = args[10]
            dhcpServerPoolSize = args[11]
        else
            interfaceName = args.interface
            hwaddr = args.hwaddr
            staticAddresses = args.staticAddresses
            dnsServers = args.dnsServers
            gateways = args.gateways
            ipv6AcceptRA = args.ipv6AcceptRA
            ipMasquerade = args.ipMasquerade
            dhcpClient = args.dhcpClient
            dhcpServer = args.dhcpServer
            dhcpServerPoolOffset = args.dhcpServerPoolOffset
            dhcpServerPoolSize = args.dhcpServerPoolSize
        end
    end
    
    if interfaceName ~= nil and type(interfaceName) ~= 'string' then
        return nil, errType.codes.InvalidParams, 'Failed to read parameter \'interface\'.'
    end
    if hwaddr ~= nil and type(hwaddr) ~= 'string' then
        return nil, errType.codes.InvalidParams, 'Failed to read parameter \'hwaddr\'.'
    end
    if staticAddresses ~= nil and utils.isArray(staticAddresses) == false then
        return nil, errType.codes.InvalidParams, 'Parameter \'staticAddresses\' must be an array.'
    end
    if dnsServers ~= nil and utils.isArray(dnsServers) == false then
        return nil, errType.codes.InvalidParams, 'Parameter \'dnsServers\' must be an array.'
    end
    if gateways ~= nil and utils.isArray(gateways) == false then
        return nil, errType.codes.InvalidParams, 'Parameter \'gateways\' must be an array.'
    end
    if checkBooleanArgument(ipv6AcceptRA) == false then
        return nil, errType.codes.InvalidParams, 'Parameter \'ipv6AcceptRA\' must have boolean value.'
    end
    if checkBooleanArgument(ipMasquerade) == false then
        return nil, errType.codes.InvalidParams, 'Parameter \'ipMasquerade\' must have boolean value.'
    end
    if checkBooleanArgument(dhcpClient) == false then
        return nil, errType.codes.InvalidParams, 'Parameter \'dhcpClient\' must have boolean value.'
    end
    if checkBooleanArgument(dhcpServer) == false then
        return nil, errType.codes.InvalidParams, 'Parameter \'dhcpServer\' must have boolean value.'
    end
    
    if dhcpClient and dhcpServer then
        return nil, errType.codes.InvalidParams, 'Parameters \'dhcpServer\' and \'dhcpClient\' cannot be set at the same time.'
    end
    
    if dhcpServerPoolOffset ~= nil and (checkInt(dhcpServerPoolOffset) == false or dhcpServerPoolOffset < 0) then
        return nil, errType.codes.InvalidParams, 'Parameter \'dhcpServerPoolOffset\' must be positive integer value.'
    end
    if dhcpServerPoolOffset ~= nil and dhcpServer ~= true then
        return nil, errType.codes.InvalidParams, 'Parameter \'dhcpServerPoolOffset\' must be set only if \'dhcpServer\' is set.'
    end
    if dhcpServerPoolSize ~= nil and (checkInt(dhcpServerPoolSize) == false or dhcpServerPoolSize < 0) then
        return nil, errType.codes.InvalidParams, 'Parameter \'dhcpServerPoolSize\' must be positive integer value.'
    end
    if dhcpServerPoolSize ~= nil and dhcpServer ~= true then
        return nil, errType.codes.InvalidParams, 'Parameter \'dhcpServerPoolSize\' must be set only if \'dhcpServer\' is set.'
    end
    
    if staticAddresses then
        res = checkIPsWithMask(staticAddresses)
        if res then
            return nil, errType.codes.InvalidParams, res
        end
    end
    
    if dnsServers then
        res = checkIPsWithoutMask(dnsServers)
        if res then
            return nil, errType.codes.InvalidParams, res
        end
    end
    
    if gateways then
        res = checkIPsWithoutMask(gateways)
        if res then
            return nil, errType.codes.InvalidParams, res
        end
    end
    
    if interfaceName == nil and hwaddr == nil then
        return nil, errType.codes.InvalidParams, 'Paramter \'interface\' or parameter \'hwaddr\' must be specified.'
    end
    if interfaceName ~= nil and hwaddr ~= nil then
        return nil, errType.codes.InvalidParams, 'Only one paramter from [\'interface\', \'hwaddr\'] must be specified.'
    end
    
    local interfaces = Network_GetState.invoke()
    if interfaces == nil then
        return nil, errType.codes.InternalError, 'Failed to query interfaces.'
    end
    
    local interface
    for i, iface in ipairs(interfaces) do
        if interfaceName then
            if iface.name == interfaceName then
                interface = iface
            end
        else
            if iface.hwaddr == hwaddr then
                interface = iface
            end
        end
    end
    if interface == nil then
        if interfaceName == nil then
            return nil, errType.codes.InvalidParams, 'Interface with MAC address \'' .. hwaddr .. '\' not found.'
        else
            return nil, errType.codes.InvalidParams, 'Interface \'' .. interfaceName .. '\' not found.'
        end
    end
    
    local conf = ''
    
    conf = conf .. '[Match]\n'
    if interfaceName ~= nil then
        conf = conf .. 'Name=' .. interfaceName .. '\n'
    else
        conf = conf .. 'MACAddress=' .. hwaddr .. '\n'
    end
    conf = conf .. '\n[Network]\n'
    
    if staticAddresses then
        for i, addr in ipairs(staticAddresses) do
            conf = conf .. 'Address=' .. addr .. '\n'
        end
    end
    if dnsServers then
        for i, dns in ipairs(dnsServers) do
            conf = conf .. 'DNS=' .. dns .. '\n'
        end
    end
    if gateways then
        for i, gw in ipairs(gateways) do
            conf = conf .. 'Gateway=' .. gw .. '\n'
        end
    end
    if ipv6AcceptRA ~= nil then
        conf = conf .. 'IPv6AcceptRA=' .. bool2ConfigValue(ipv6AcceptRA) .. '\n'
    end
    if ipMasquerade ~= nil then
        conf = conf .. 'IPMasquerade=' .. bool2ConfigValue(ipMasquerade) .. '\n'
    end
    if dhcpClient ~= nil then
        if dhcpClient then
            conf = conf .. 'DHCP=yes\n'
        else
            conf = conf .. 'DHCP=no\n'
        end
    end
    if dhcpServer ~= nil then
        conf = conf .. 'DHCPServer=' .. bool2ConfigValue(dhcpServer) .. '\n'
        if dhcpServerPoolSize or dhcpServerPoolOffset then
            conf = conf .. '\n[DHCPServer]\n'
            if dhcpServerPoolOffset then
                conf = conf .. 'PoolOffset=' .. dhcpServerPoolOffset .. '\n'
            end
            if dhcpServerPoolSize then
                conf = conf .. 'PoolSize=' .. dhcpServerPoolSize .. '\n'
            end
        end
    end
    
    local fileName
    if interfaceName then
        fileName = interfaceName .. '.network'
    else
        fileName = interfacesService.hwaddr2name(hwaddr) .. '.network'
    end
    
    local filesystemLocked = false
    if utils.isFilesystemReadOnly() then
        filesystemLocked = true
        utils.unlockFilesystem()
    end
    
    local res, err, errMsg = utils.writeToFile('/etc/systemd/network/' .. fileName, conf)
    if res ~= nil then
        return nil, err, errMsg
    end
    
    local oldConfig
    if interfaceName then
        oldConfig = '/etc/systemd/network/' .. interfacesService.hwaddr2name(interface.hwaddr) .. '.network'
    else
        oldConfig = '/etc/systemd/network/' .. interface.name .. '.network'
    end
    if utils.fileExists(oldConfig) then
        local ret, err, msg = removeMethod.invoke({oldConfig})
        if err ~= nil then
            return nil, err, 'Failed to remove old configuration file ' .. oldConfig .. '.\n' .. msg
        end
    end
    
    flushInterface(interface.name)
    
    res = restartNetworkService()
    
    if filesystemLocked then
        utils.lockFilesystem()
    end
    
    if res == false then
        return nil, errType.codes.InternalError, 'Failed to restart network service.'
    end
end

return Network_SetConfig
