local errType = require 'errType'
local utils = require 'utils'
local cjson = require 'cjson.safe'

local Syslog_GetState = {}

local MAX_LOG_COUNT = 500;

function Syslog_GetState.invoke(args)
    local unit, sinceTime, untilTime, count
    
    if args ~= nil then
        if utils.isArray(args) then
            unit = args[1]
            sinceTime = args[2]
            untilTime = args[3]
            count = args[4]
        else
            unit = args.unit
            sinceTime = args.sinceTime
            untilTime = args.untilTime
            count = args.count
        end
    end
    
    if unit ~= nil and type(unit) ~= 'string' then
        return nil, errType.codes.InvalidParams, 'Failed to read parameter \'unit\'.'
    end
    if sinceTime ~= nil and type(sinceTime) ~= 'string' then
        return nil, errType.codes.InvalidParams, 'Failed to read parameter \'sinceTime\'.'
    end
    if untilTime ~= nil and type(untilTime) ~= 'string' then
        return nil, errType.codes.InvalidParams, 'Failed to read parameter \'untilTime\'.'
    end
    if count ~= nil and utils.isInt(count) == false then
        return nil, errType.codes.InvalidParams, 'Failed to read parameter \'count\'.'
    end
    
    local argv = {'journalctl', '--no-pager'}
    if count ~= nil and count < MAX_LOG_COUNT then
        table.insert(argv, '-n')
        table.insert(argv, count)
    else
        table.insert(argv, '-n')
        table.insert(argv, MAX_LOG_COUNT)
    end
    table.insert(argv, '-o')
    table.insert(argv, 'json')
    
    if unit ~= nil then
        table.insert(argv, '-u')
        table.insert(argv, unit)
    end
    if sinceTime ~= nil then
        table.insert(argv, '--since')
        table.insert(argv, sinceTime)
    end
    if untilTime ~= nil then
        table.insert(argv, '--until')
        table.insert(argv, untilTime)
    end
    
    local stdout, err, errMsg = utils.shell(argv)
    if utils.trim(errMsg) == 'Failed to get realtime timestamp: Cannot assign requested address' then
        return cjson.empty_array
    end
    if err ~= nil then
        return nil, err, errMsg
    end
    
    local logs = {}
    
    local lines = utils.split(stdout, '\n')
    for key, line in pairs(lines) do
        local fullLog = cjson.decode(line)
        if fullLog ~= nil then
            local logItem = {}
            logItem.unit = fullLog._SYSTEMD_UNIT
            if logItem.unit == nil then
                logItem.unit = fullLog._TRANSPORT
            end
            logItem.msg = fullLog.MESSAGE
            logItem.exe = fullLog._EXE
            logItem.priority = fullLog.PRIORITY
            logItem.timestamp = fullLog.__REALTIME_TIMESTAMP
            logItem.time = os.date('%Y-%m-%d %H:%M:%S', string.sub(logItem.timestamp, 1, 10))
            logItem.timestamp = tonumber(logItem.timestamp)
            table.insert(logs, logItem)
        end
    end
    utils.reverseTable(logs)

    if #logs > 0 then
        return logs
    else
        return cjson.empty_array
    end
end

return Syslog_GetState
