local cjson = require 'cjson.safe'

local utils = require 'utils'
local errType = require 'errType'

local echo_method = require 'echo'
local Disks_GetState_method = require 'Disks_GetState'
local RamUsage_GetState_method = require 'RamUsage_GetState'
local TimeDate_GetState_method = require 'TimeDate_GetState'
local TimeDate_SetConfig_method = require 'TimeDate_SetConfig'
local LoadAverage_GetState_method = require 'LoadAverage_GetState'
local Hostname_GetState_method = require 'Hostname_GetState'
local Hostname_SetConfig_method = require 'Hostname_SetConfig'
local Cpu_GetState_method = require 'Cpu_GetState'
local alternativeExecutivesService = require 'alternativeExecutives'
local RexCore_SetConfig_method = require 'RexCore_SetConfig'
local RexCore_GetState_method = require 'RexCore_GetState'
local System_Shutdown_method = require 'System_Shutdown'
local FileSystem_List_method = require 'FileSystem_List'
local FileSystem_GetFileContent_method = require 'FileSystem_GetFileContent'
local FileSystem_SetFileContent_method = require 'FileSystem_SetFileContent'
local FileSystem_Move_method = require 'FileSystem_Move'
local FileSystem_Copy_method = require 'FileSystem_Copy'
local FileSystem_CreateDirectory_method = require 'FileSystem_CreateDirectory'
local FileSystem_CreateFile_method = require 'FileSystem_CreateFile'
local FileSystem_Remove_method = require 'FileSystem_Remove'
local Certificate_Read_method = require 'Certificate_Read'
local Certificate_Write_method = require 'Certificate_Write'
local Network_GetState_method = require 'Network_GetState'
local Network_SetConfig_method = require 'Network_SetConfig'
local Network_RemoveConfig_method = require 'Network_RemoveConfig'
local Syslog_GetState_method = require 'Syslog_GetState'
local Meech_GetUsers_method = require 'Meech_GetUsers'
local Meech_SetUsers_method = require 'Meech_SetUsers'

local methods = {
    echo = echo_method,
    Disks_GetState = Disks_GetState_method,
    RamUsage_GetState = RamUsage_GetState_method,
    TimeDate_GetState = TimeDate_GetState_method,
    LoadAverage_GetState = LoadAverage_GetState_method,
    TimeDate_SetConfig = TimeDate_SetConfig_method,
    Hostname_GetState = Hostname_GetState_method,
    Hostname_SetConfig = Hostname_SetConfig_method,
    Network_GetState = Network_GetState_method,
    Cpu_GetState = Cpu_GetState_method,
    alternativeExecutives = alternativeExecutivesService,
    RexCore_GetState = RexCore_GetState_method,
    RexCore_SetConfig = RexCore_SetConfig_method,
    System_Shutdown = System_Shutdown_method,
    FileSystem_List = FileSystem_List_method,
    FileSystem_GetFileContent = FileSystem_GetFileContent_method,
    FileSystem_SetFileContent = FileSystem_SetFileContent_method,
    FileSystem_Move = FileSystem_Move_method,
    FileSystem_Copy = FileSystem_Copy_method,
    FileSystem_CreateDirectory = FileSystem_CreateDirectory_method,
    FileSystem_CreateFile = FileSystem_CreateFile_method,
    FileSystem_Remove = FileSystem_Remove_method,
    Certificate_Read = Certificate_Read_method,
    Certificate_Write = Certificate_Write_method,
    Network_SetConfig = Network_SetConfig_method,
    Network_RemoveConfig = Network_RemoveConfig_method,
    Syslog_GetState = Syslog_GetState_method,
    Meech_GetUsers = Meech_GetUsers_method,
    Meech_SetUsers = Meech_SetUsers_method
}

local function createErrorResponse(id, errCode)
    local resp = {
        jsonrpc = '2.0',
        error = {
            code = errCode,
            message = errType.getMsgFromCode(errCode)
        },
        id = id
    }
    
    if id == nil then
        resp.id = cjson.null
    end
    
    return resp
end

local function createErrorResponseForRequest(id, errCode, errMsg)
    -- notification
    if id == nil then
        return nil
    end
    
    local resp = createErrorResponse(id, errCode)
    resp.id = id;
    
    if errMsg ~= nil then
        resp.error.data = {}
        resp.error.data.message = errMsg
    end
    
    return resp
end

local function checkParams(params)
    if params == nil then
        return true
    end
    
    if type(params) == 'table' then
        return true
    end
    
    return false
end

local function checkRequest(req)
    
    if req.method == nil or type(req.method) ~= 'string' then
        return false
    end
    
    if req.jsonrpc == nil or type(req.jsonrpc) ~= 'string' or req.jsonrpc ~= '2.0' then
        return false
    end
    
    return true
end

local function handleRequest(req)
    
    if type(req) ~= 'table' then
        return createErrorResponse(nil, errType.codes.InvalidRequest)
    end
    
    if checkRequest(req) == false then
        return createErrorResponse(req.id, errType.codes.InvalidRequest)
    end
    
    if checkParams(req.params) == false then
        return createErrorResponse(req.id, errType.codes.InvalidRequest)
    end
    
    local methodArg = req.method
    
    for key, method in pairs(methods) do
        if key == methodArg then
            local res, err, errMsg = method.invoke(req.params)
            
            if err ~= nil then
                return createErrorResponseForRequest(req.id, err, errMsg)
            else
                -- notification
                if req.id == nil then
                    return nil;
                end
                
                local result = '';
                if res ~= nil then
                    result = res
                end
                return {
                    jsonrpc = '2.0',
                    id = req.id,
                    result = result
                }
            end
        end
    end
    
    return createErrorResponseForRequest(req.id, errType.codes.MethodNotFound, nil)
end





ngx.req.read_body()

if ngx.req.get_method() ~= 'POST' then
    ngx.exit(ngx.HTTP_NOT_ALLOWED)
end

local buf = ngx.req.get_body_data()

--buf == nil => no body or body was stored in temp file
if buf == nil then 
    local fileName = ngx.req.get_body_file()
    local file = io.open(fileName, "rb")
    if not file then
        return nil
    end
    buf = file:read("*a")
    file:close()
end

if buf == nil then
    local resp = createErrorResponse(nil, errType.codes.InvalidRequest)
    ngx.say(cjson.encode(resp))
    ngx.exit(ngx.HTTP_OK)
end

local body = cjson.decode(buf)

if body == nil or type(body) ~= 'table' then
    local resp = createErrorResponse(nil, errType.codes.ParseError)
    ngx.say(cjson.encode(resp))
    ngx.exit(ngx.HTTP_OK)
end

if utils.isArray(body) then
    
    if #body == 0 then
        local resp = createErrorResponse(nil, errType.codes.InvalidRequest)
        ngx.say(cjson.encode(resp))
        ngx.exit(ngx.HTTP_OK)
    end
    
    -- MULTITHREADED APPROACH
    -- local threads = {}
    -- for k, req in pairs(body) do
    --     local thread = ngx.thread.spawn(handleRequest, req)
    --     table.insert(threads, thread)
    -- end
    -- local responses = {}
    -- for i = 1, #threads do
    --     local ok, res = ngx.thread.wait(threads[i])
    --     if ok then
    --         if res ~= nil then
    --             table.insert(responses, res)
    --         end
    --     else
    --         local req = body[i]
    --         if type(req) == 'table' then
    --             table.insert(responses, createErrorResponseForRequest(req.id, errType.codes.ServerError, nil))
    --         else
    --             table.insert(responses, createErrorResponse(nil, errType.codes.ServerError))
    --         end
    --     end
    -- end

    local responses = {}
    
    for k, req in pairs(body) do
        local response = handleRequest(req)
        if response then
            table.insert(responses, response)
        end
    end
    
    if #responses > 0 then
        ngx.say(cjson.encode(responses))
    end
    
    ngx.exit(ngx.HTTP_OK)
else
    local response = handleRequest(body)
    if response ~= nil then
        ngx.say(cjson.encode(response))
    end
    ngx.exit(ngx.HTTP_OK)
end
